#!/usr/bin/perl
#Rene Warren 2006,2007
#Short Sequence Assembly by Kmer search and 3' Extension (SSAKE)

#NAME
#   SSAKE v1.2  Rene Warren, September 2007

#SYNOPSIS
#   Progressive clustering of millions of short DNA sequences by Kmer extension

#LICENSE
#   Copyright (c) 2006-2007 Canada's Michael Smith Genome Science Centre.  All rights reserved.

#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the GNU General Public License
#   as published by the Free Software Foundation; either version 2
#   of the License, or (at your option) any later version.

#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.


use strict;
use Benchmark;
require "getopts.pl";
use vars qw($opt_f $opt_m $opt_s $opt_v $opt_t);
&Getopts('f:m:l:s:v:t:');
my ($stringent,$max_trim,$min_overlap,$verbose,$MIN_READ_LENGTH,$SEQ_SLIDE)=(0,0,11,0,11,1);
my $per;

if(! $opt_f){
   print "Usage: $0\n";
   print "-f <A single fasta file containing all the sequence reads (.fa file)>\n";
   print "-m <Minimum word length used for k-mer search and sequence read overlap (default -m 11)>\n";
   print "-t <up to -t number of bases to trim on the contig end when all possibilities have been exhausted for an extension (default -t 0)>\n";
   print "-s <Runs in stringent mode.  When set, this option will prevent contig extension if more than one sequence read has a matching k-mer (1=yes 0=no, default -s 0)>\n";
   die "-v <Runs in Verbose mode (1=yes 0=no, default -v 0)> (optional)\n";
}

my $file = $opt_f;

$min_overlap = $opt_m if ($opt_m);
$stringent = $opt_s if ($opt_s);
$max_trim = $opt_t if ($opt_t);
$verbose = $opt_v if ($opt_v);

if(! -e $opt_f){
   die "Invalid file $opt_f.\n";
}


my $pid = getpgrp(0);
my $contig = $file . "." . $pid .  ".contigs";
my $singlet = $file . "." . $pid . ".singlets";
my $short = $file . "." . $pid . ".short";
my $log = $file . "." . $pid . ".log";

open (LOG, ">$log") || die "can write to $log..Exiting.\n";;

if($min_overlap < 11 || $min_overlap > 50){
   print LOG "-m must be a number between 11-50..Exiting.\n";
   close LOG;
   exit;
}


my $init_message = "\nRunning: $0\n-f $file\n-m $min_overlap\n-s $stringent\n-t $max_trim\nContigs file: $contig\nSinglets file: $singlet\nExcluded short reads file: $short\nLog file: $log\n";

print $init_message;
print LOG $init_message;

my $t0 = new Benchmark;
my $date = `date`;
chomp($date);

my $reading_reads_message = "\n\nReading sequences initiated $date\n";
print $reading_reads_message;
print LOG $reading_reads_message;
my ($set,$bin) = &readFile($file);

my $t1 = new Benchmark;
$date = `date`;
chomp($date);

my $ssake_start_message = "\n\nSequence assembly initiated $date\n";
print $ssake_start_message;
print LOG $ssake_start_message;

#-------------------------------------------------
my ($tig_count,$previous_index) = (0,0);

open (TIG, ">$contig") || die "can write to $contig. Exiting.\n";
open (SIN, ">$singlet") || die "can write to $singlet. Exiting.\n";
open (SHO, ">$short") || die "can write to $short. Exiting.\n";

eval{

my $status_bar = "+";
for(my $i=1;$i<=99;$i++){
   $per->{$i}++;
   my $ct = $i /10;
   if($ct == int($ct)){$status_bar .= $ct;}else{$status_bar .= "-";}
}
$status_bar .= "+ x 10 [% complete]";
print "$status_bar\n.";

my $keys_start = keys ( %$set );

ASSEMBLY:
foreach my $seq (sort {$set->{$b}<=>$set->{$a}} keys %$set){#cycle through the input [normalized] reads

   my $start_sequence = $seq;
   my $reads_needed = $set->{$seq};                          #tracks coverage
   my $orig_mer = length($seq);
   my $total_bases = $orig_mer;

   if(defined $set->{$seq} && $orig_mer > $MIN_READ_LENGTH && $orig_mer > $min_overlap){#sequence read hasn't been used, is longer than 11 nt and the user-defined overlap minimum -m
      #### Delete keys ref
      my @o=split(//,$seq);                               

      $_ = $seq;
      tr/ACTG/TGAC/;
      my $comp_seq = reverse();
 
      my @c=split(//,$comp_seq);
                                                            #remove kmer from hash table and prefix tree
      delete $bin->{$o[0]}{$o[1]}{$o[2]}{$o[3]}{$o[4]}{$o[5]}{$o[6]}{$o[7]}{$o[8]}{$o[9]}{$o[10]}{$seq};
      delete $bin->{$c[0]}{$c[1]}{$c[2]}{$c[3]}{$c[4]}{$c[5]}{$c[6]}{$c[7]}{$c[8]}{$c[9]}{$c[10]}{$comp_seq};
      delete $set->{$seq};
      ####
     
      print "\n>>>Start Sequence:$seq<<<\n" if ($verbose);

      #3'extension
      my ($ct,$pos,$possibility,$trim_counter,$current_reads,$current_bases) = (0,0,0,0,0,0);
      my $span = $orig_mer;
      my $previous;

      #####################
      #Decrease the contig by one base
      while($trim_counter <= $max_trim){

         EXT3:
         while ($span >= $min_overlap){                                                                                #will slide the subseq, until the user-defined min overlap size
            $ct += $SEQ_SLIDE;
            $span = $orig_mer - $ct;
            $pos = length($seq) - $span; 

            my $subseq = substr($seq, $pos, $span);                                                            #make a sub-sequence of length l-(1..i) for searching

            my @s=split(//,$subseq);
            my $subset = $bin->{$s[0]}{$s[1]}{$s[2]}{$s[3]}{$s[4]}{$s[5]}{$s[6]}{$s[7]}{$s[8]}{$s[9]}{$s[10]}; #Will grab everything even the reverse complement ones

            print "####3' EXTENSION Counter:$ct .. Span:$span .. Subseq:$subseq Previous:$previous\n" if ($verbose);

            $possibility = 0;
            $previous = $seq;         

            SEARCH3:                                                                                          #this cycles through limited kmer space
            foreach my $pass (sort {$subset->{$b} <=> $subset->{$a}} keys %$subset){
  
               if($pass =~ /^$subseq([ACGT]*)/){                                                              #can we align perfectly that subseq kmer to another rd start?
  
                  my $dangle = $1;
                  $possibility++;
   
                  last EXT3 if ($possibility > 1 && $stringent==1);                                           #search ends if stringent setting and many possibilities 
  
                  #This is the contig extension
                  $seq .= $dangle;                                                                                      
                  $trim_counter=0;

                  $ct = 0;
                  print "3' SeqRd:$pass -> Subset:$subseq -> Added:$dangle\n\nConcat:$seq\n\n" if ($verbose);

                  #Will delete rc counterpart
                  $_ = $pass;
                  tr/ACTG/TGAC/;
                  my $complement=reverse();

                  if(defined $set->{$pass}){
                     $current_reads = $set->{$pass};
                     $current_bases = length($pass) * $current_reads;
                     $reads_needed += $current_reads;
                     $total_bases += $current_bases;	
                     delete $set->{$pass};
                  }
                  if(defined $set->{$complement}){
                     $current_reads = $set->{$complement};
                     $current_bases = length($complement) * $current_reads;
                     $reads_needed += $current_reads;
                     $total_bases += $current_bases;
                     delete $set->{$complement};
                  }

                  delete $bin->{$s[0]}{$s[1]}{$s[2]}{$s[3]}{$s[4]}{$s[5]}{$s[6]}{$s[7]}{$s[8]}{$s[9]}{$s[10]}{$pass};
                  my @r=split(//,$complement);
                  delete $bin->{$r[0]}{$r[1]}{$r[2]}{$r[3]}{$r[4]}{$r[5]}{$r[6]}{$r[7]}{$r[8]}{$r[9]}{$r[10]}{$complement};
                  last SEARCH3 if (! $stringent);                                                           #search ends when first possibility encountered (non stringent)
               }
            }
         }#while overlap >= user-defined -m minimum

         $trim_counter++;
         ($ct,$pos,$possibility) = (0,0,0);
         $span = $orig_mer;
         $seq = substr($seq, 0, -1) if ($trim_counter <= $max_trim);
         print "3'EXTENSION - NOTHING TO BE DONE IN 3', UNLESS TRIM $trim_counter nt UNTIL $max_trim\nTRIMMED CONTIG :$seq\n\n" if ($verbose);

      }#Loops through base removal on end
      ############

      if($possibility > 1 && $stringent==1){   ###revert to previous contig if not allowing multiple path
         print "3' EXTENSION FORCE-STOPPED DUE TO LIMITING VALUE OF -s $stringent AND NUMBER OF POSSIBLE EXTENSIONS (>1) - WILL REVERT FROM\n$seq\nto\n$previous\n" if ($verbose);
         $seq = $previous;               #revert to previous sequence, before extension
         $reads_needed -= $current_reads;
         $total_bases -= $current_bases;
      }

      print "\n\n*** NOTHING ELSE TO BE DONE IN 3' PERHAPS YOU COULD DECREASE THE MIN. KMER AND INCREASE THE SLIDE (Set to $min_overlap and $SEQ_SLIDE, respectively) ***\n\n" if ($verbose);
      ####this marks the end of the 3' extension, and the beginning of the 5'

      ####Reverse-complement the contig in order to change the 5' extension problem into a 3' extension
      $_ = $seq;
      tr/ACTG/TGAC/;
      my $seqrc = reverse();
      my $rctig = $seqrc;
      my $previousrc = '';
      #####

      ($ct,$pos,$possibility,$trim_counter,$current_reads,$current_bases) = (0,0,0,0,0,0);
      $span = $orig_mer;

      #####################
      #Decrease the contig by one
      while($trim_counter <= $max_trim){

         EXT5:
         while ($span >= $min_overlap){                                #will slide the subseq, until the min kmer size

            $ct += $SEQ_SLIDE;
            $span = $orig_mer - $ct;
            $pos = length($seqrc) - $span;

            my $subseq = substr($seqrc, $pos, $span);
            my @s=split(//,$subseq);
            my $subset = $bin->{$s[0]}{$s[1]}{$s[2]}{$s[3]}{$s[4]}{$s[5]}{$s[6]}{$s[7]}{$s[8]}{$s[9]}{$s[10]};
            print "####5' EXTENSION through 3' -- Counter:$ct .. Span:$span .. Subseq:$subseq\n" if ($verbose);

            ###To count possible path and revert to original seq
            $possibility = 0;
            $previousrc = $seqrc;

            SEARCH5:
            foreach my $pass(sort {$subset->{$b} <=> $subset->{$a}} keys %$subset){ ##sort helps cluster high coverage areas first

               if($pass =~ /^$subseq([ACGT]*)/){ #can we align perfectly that subseq kmer to another rd start?
                  my $dangle = $1;
                  $possibility++;
                  last EXT5 if ($possibility >1 && $stringent==1);

                  #This is the contig extension
                  $seqrc .= $dangle;
                  $trim_counter = 0; ## resets the trim counter if an extension was made

                  $ct = 0;
                  print "5' SeqRd:$pass -> Subset:$subseq -> Added:$dangle\n\nConcat:$seqrc\n\n" if ($verbose);

                  #Try and see whether we've hit a rverse-complemented (rc) read
                  $_ = $pass;
                  tr/ACTG/TGAC/;
                  my $complement=reverse();
              
                  if(defined $set->{$pass}){
                     $current_reads = $set->{$pass};
                     $current_bases = length($pass) * $current_reads;
                     $reads_needed += $current_reads;
                     $total_bases += $current_bases;
                     delete $set->{$pass};
                  }
                  if(defined $set->{$complement}){
                     $current_reads = $set->{$complement};
                     $current_bases = length($complement) * $current_reads;
                     $reads_needed += $current_reads;
                     $total_bases += $current_bases;
                     delete $set->{$complement};
                  }
               
                  delete $bin->{$s[0]}{$s[1]}{$s[2]}{$s[3]}{$s[4]}{$s[5]}{$s[6]}{$s[7]}{$s[8]}{$s[9]}{$s[10]}{$pass};
                  my @r=split(//,$complement);
                  delete $bin->{$r[0]}{$r[1]}{$r[2]}{$r[3]}{$r[4]}{$r[5]}{$r[6]}{$r[7]}{$r[8]}{$r[9]}{$r[10]}{$complement};

                  last SEARCH5 if (! $stringent);
               }
            }
         }

         $trim_counter++;
         ($ct,$pos,$possibility) = (0,0,0);
         $span = $orig_mer;
         $seqrc = substr($seqrc, 0, -1) if ($trim_counter <= $max_trim);
         print "5' EXTENSION - NOTHING TO BE DONE IN 5', UNLESS TRIM $trim_counter nt UNTIL $max_trim\nTRIMMED CONTIG :$seqrc\n\n" if ($verbose);

      }#Loop through base removal on end
      ############

      if($possibility > 1 && $stringent==1){   ###revert to previous contig if not allowing multiple path
         print "5' EXTENSION FORCE-STOPPED DUE TO LIMITING VALUE OF -s $stringent AND NUMBER OF POSSIBLE EXTENSIONS (>1) - WILL REVERT FROM\n$seqrc\nto\n$previousrc\n" if ($verbose);
         $seqrc = $previousrc;
         $reads_needed -= $current_reads;
         $total_bases -= $current_bases;
      }

      $tig_count++;
      my $leng = length($seqrc);
      my $cov =  $total_bases / $leng;

      $_ = $seqrc;
      tr/ACTG/TGAC/;
      my $reversetig = reverse();

      if($start_sequence ne $seqrc && $start_sequence ne $reversetig){
         printf TIG ">contig%i|size%i|read%i|cov%.2f\n%s\n", ($tig_count,$leng,$reads_needed,$cov,$reversetig);    #print contigs to file
      }else{
         printf SIN ">contig%i|size%i|read%i|cov%.2f\n%s\n", ($tig_count,$leng,$reads_needed,$cov,$start_sequence);    #print singlets to file
      }

   }else{ 
      if($orig_mer < $MIN_READ_LENGTH){
         print SHO "$seq\tInput sequence shorter than minimum read length allowed ($orig_mer < $MIN_READ_LENGTH nt)\n";
      }elsif($orig_mer < $min_overlap){
         print SHO "$seq\tInput sequence shorter than minimum overlap specified($orig_mer < -m $min_overlap)\n";
      }
   }

   my $keys_left = keys( %$set );
   my $index = (int((($keys_start-$keys_left)/$keys_start)*100));
   if(defined $per->{$index}){
      print "." x ($index - $previous_index);
      $|=1; ###clear buffer
      delete $per->{$index};
   }
   $previous_index = $index;

   last ASSEMBLY if (! $keys_left);
}
print ".";
};###end eval block

my $t2 = new Benchmark;
$date = `date`;
chomp($date);

if($@){
   my $message = $@;
   my $failure = "\n\nSomething went wrong running $0 $date\n$message\n\n";
   print $failure;
   print LOG $failure; 
}else{
   my $success = "\n\n$0 executed normally $date\n";
   print $success;
   print LOG $success;
}

my $read_time = timediff($t1,$t0);
my $assembly_time = timediff($t2,$t1);
my $total_time = timediff($t2,$t0);

printf LOG "\nTime to read the input file: %s \nTime to assemble: %s\nTotal run time: %s", (timestr($read_time), timestr($assembly_time), timestr($total_time));

close TIG;
close SIN;
close LOG;
close SHO;

exit;


#-----------------
sub readFile{
   my $file =shift;

   my ($set,$bin);

   my $ctrd=0;

   open(IN,$file);
   while(<IN>){
      chomp;
      if(/^([ACGT]*)$/i){
         
         my $seq=$1;
         my $orig=uc($seq);  

         if ($orig ne ''){ 
            ####show progress
            my $s10k = $ctrd / 10000;
            print "." if ($s10k == int($s10k) && $s10k && $ctrd);
            $|=1; ###clear buffer
             
            my $s100k = $ctrd / 100000;
            printf "%i sequences inputted\n", ($ctrd) if ($s100k == int($s100k) && $s100k && $ctrd);

            my @sf=split(//,$orig);
            $set->{$orig}++;

            $_ = $orig;
            tr/ACTG/TGAC/;
            my $rc=reverse();
            
            my @sr=split(//,$rc);

            $bin->{$sf[0]}{$sf[1]}{$sf[2]}{$sf[3]}{$sf[4]}{$sf[5]}{$sf[6]}{$sf[7]}{$sf[8]}{$sf[9]}{$sf[10]}{$orig}++;
            $bin->{$sr[0]}{$sr[1]}{$sr[2]}{$sr[3]}{$sr[4]}{$sr[5]}{$sr[6]}{$sr[7]}{$sr[8]}{$sr[9]}{$sr[10]}{$rc}++;
            $ctrd++;
         }
      }
   }
   my $read_number_message = "$ctrd total sequences (" . keys( %$set ) . " unique)\n";
   printf $read_number_message;
   print LOG $read_number_message;

   close IN;
   return $set,$bin;
}
